#include<NoDelay.h>
#include <Servo.h>
#include <EEPROM.h>


// I/O Pins definitons
#define outsideTempPin A0 // Outside Tempature Sensor Pin
#define caseTempPin A1 // Case Temparature Sensor Pin
#define insideTempPin A2 // Room Temparature Sensor Pin
#define heatingControlPin 3 // Resistive Heating Control Pin
#define fanControlPin 9 // Fan Control Pin
#define servoPin 8 // Servo Pin
#define buttonPin 2 // Push Button Pin (interrupts)

// Constants for resistive heating
const float dryAirSpecificHeat = 1.00; // J/gK
const float dryAirDensity = 1275; // g/m^3

// Parameters for resistive heating
float roomHeight = 0.71; // m
float roomArea = 0.2706; // m^2
float sourceVoltage = 5; // voltage
float resistorRating = 0.25; // watt
int resistanceValue = 220; // ohm
int heatingTime = 20; // minutes
int numberResistance = 17;

// Tempature Variables
float targetTemp = 25; // celcius
float temperatureThreshold = 1; // celcius
float heatingThreshold = 3; // celcius
float heatingFanThreshold = 2; // celcius
// Arduino Variables
float outsideTemp;
float caseTemp;
float insideTemp; 
//  Servo Variables
Servo servo;
bool directionState = false; // FALSE = Intake is Open, TRUE = Intake is Closed (Default State = Intake is Open)

//Volatile variables
bool buttonPressed = false;
float currentVoltage;
int pos = 0;

//NoDelay setup to call system update function every 5 seconds
void UpdateSystem(); // Function declaration before NoDelay definition
noDelay sensorDelay(10000, UpdateSystem);
void setup()
{
    buttonPressed = false;
    GetServoStateFromMemory();
    servo.attach(servoPin);
    pinMode(outsideTempPin, INPUT);
    pinMode(caseTempPin, INPUT);
    pinMode(insideTempPin, INPUT);
    pinMode(heatingControlPin, OUTPUT);
    pinMode(fanControlPin, OUTPUT);
    pinMode(buttonPin, INPUT);
    attachInterrupt(digitalPinToInterrupt(buttonPin), buttonToggle, INPUT_PULLUP); // Triggers interrupt when button goes from HIGH to LOW (So 
    analogWrite(heatingControlPin, 0);
    analogWrite(fanControlPin, 0);
    Serial.begin(9600);
}

void GetServoStateFromMemory() {
  bool getState = false;
  EEPROM.get(0, directionState); //Read from memory to get the last state of intake
  if(isnan(directionState)) directionState = false; //Set to default (intake is open state)
}

void loop()
{
  if(buttonPressed) buttonToggle();
  sensorDelay.update();
}

void UpdateSystem() {
  outsideTemp = analogRead(outsideTempPin) * 500.0 / 1024 - 60;
  caseTemp = analogRead(caseTempPin) * 500.0 / 1024 - 60;
  insideTemp =analogRead(insideTempPin) * 500.0 / 1024 - 60;

  if( abs(targetTemp - insideTemp) <= temperatureThreshold) { // Tempature has reached acceptable threshold
    // Standy By Mode
    //analogWrite(fanControlPin, 126);
    //analogWrite(heatingControlPin,0);
    Serial.println("The room has reached optimal tempature.");
    Serial.println("Stand by mode activated.\n");
  }
  else if(targetTemp > insideTemp) {
    HeatingMode();
  }
  else {
    CoolingMode();
  }

  Serial.print("\n\nThe outside temparature is now: ");
  Serial.print(outsideTemp);
  Serial.print(" ºC\n");

  Serial.print("The case temparature is now: ");
  Serial.print(caseTemp);
  Serial.print(" ºC\n");

  Serial.print("The inside temparature is now: ");
  Serial.print(insideTemp);
  Serial.print(" ºC\n");

  Serial.print("The targete temparature is now: ");
  Serial.print(targetTemp);
  Serial.print(" ºC\n\n"); }

void CoolingMode() {
  // Next step is to identify how the case temp, outisde temp impact the inside temparature
  float fanValue;
  if(directionState) {
    if( abs(targetTemp - caseTemp) <= (temperatureThreshold/2)) { //Closed Intake Cooling Case 1 : abs(Target Temperature - Case Temperature) <= (temperatureThreshold/2)
      Serial.println("Closed Intake Cooling Case 1 : abs(Target Temperature - Case Temperature) <= (temperatureThreshold/2)");
      fanValue = 255; // Case Temp is ideal temperature (Fan at max output)
    }
    else if(targetTemp < caseTemp) {  //Closed Intake Cooling Case 2 : (Target Temperature < Case Temperature)
      Serial.println("Closed Intake Cooling Case 2 : (Target Temperature < Case Temperature)");
      fanValue = 255 - mapFloat(abs(targetTemp - caseTemp),0,10,1,255); // Higher case temperature means lower fan output
    }
    else { //Closed Intake Cooling Case 3 : (Target Temperature > Case Temperature)
      Serial.println("Closed Intake Cooling Case 3 : (Target Temperature > Case Temperature)");
      fanValue = mapFloat(abs(targetTemp - caseTemp),0,10,1,255); // Lower case temperature means higher fan output 
    } 
  }
  else {
    if( abs(targetTemp - outsideTemp) <= (temperatureThreshold/2)) { //Closed Intake Cooling Case 1 : abs(Target Temperature - Outside Temperature) <= (temperatureThreshold/2)
      Serial.println("Closed Intake Cooling Case 1 : abs(Target Temperature - Outside Temperature) <= (temperatureThreshold/2)");
      fanValue = 255; // Outside Temp is ideal temperature (Fan at max output)
    }
    else if(targetTemp < outsideTemp) {  //Closed Intake Cooling Case 2 : (Target Temperature < Outside Temperature)
      Serial.println("Closed Intake Cooling Case 2 : (Target Temperature < Outside Temperature)");
      fanValue = 255 - mapFloat(abs(targetTemp - outsideTemp),0,10,1,255); // Higher outside temperature means lower fan output
    }
    else { //Closed Intake Cooling Case 3 : (Target Temperature > Outside Temperature)
      Serial.println("Closed Intake Cooling Case 3 : (Target Temperature > Outside Temperature)");
      fanValue = mapFloat(abs(targetTemp - outsideTemp),0,10,1,255); // Lower outside temperature means higher fan output 
    } 
  }
  Serial.println("Cooling Fan Value: ");
  Serial.println(fanValue);
  //analogWrite(heatingControlPin, 0); // Turn OFF resistive heating
  //analogWrite(fanControlPin, fanValue);
  Serial.println("The cooling process will now be underway");
}
 
void HeatingMode() {

  if ( ((targetTemp - insideTemp) > heatingThreshold) || (directionState && (targetTemp - caseTemp) > temperatureThreshold) || (!directionState && (outsideTemp - insideTemp) > temperatureThreshold) ) { // Tempature needs more heating
    // Additional Heating Mode
    float voltage = getVal(insideTemp,targetTemp);
    if(voltage == -1) {
        Serial.println("The target tempature is too high. It will be readjusted to the capicity of THEC.");
        targetTemp = getTargetTemp(insideTemp,5.0);
        Serial.print("The target tempature is now: ");
        Serial.print(targetTemp);
        Serial.print("ºC\n");
        voltage = 5;
    }
    //analogWrite(fanControlPin, 255);
    //analogWrite(heatingControlPin, mapFloat(voltage, 0, 5, 0, 255));
    Serial.println("The resistive heating process will be underway.");
    Serial.print("The required voltage is ");
    Serial.print(voltage);
    Serial.println(" V\n");
  }
  else { // Inside Temparature can be heated using THEC without resistive heating
    float fanValue;
    if(directionState) {
      if( abs(targetTemp - caseTemp) <= (temperatureThreshold/2)) { //Closed Intake Cooling Case 1 : abs(Target Temperature - Case Temperature) <= (temperatureThreshold/2)
        Serial.println("Closed Intake Cooling Case 1 : abs(Target Temperature - Case Temperature) <= (temperatureThreshold/2)");
        fanValue = 255; // Case Temp is ideal temperature (Fan at max output)
      }
      else if(targetTemp < caseTemp) {  //Closed Intake Cooling Case 2 : (Target Temperature < Case Temperature)
        Serial.println("Closed Intake Cooling Case 2 : (Target Temperature < Case Temperature)");
        fanValue = mapFloat(abs(targetTemp - caseTemp),0,10,1,255); // Higher case temperature means higher fan output
      }
      else { //Closed Intake Cooling Case 3 : (Target Temperature > Case Temperature)
        Serial.println("Closed Intake Cooling Case 3 : (Target Temperature > Case Temperature)");
        fanValue = 255 - mapFloat(abs(targetTemp - caseTemp),0,10,1,255); // Lower case temperature means lower fan output 
      } 
    }
    else {
      if( abs(targetTemp - outsideTemp) <= (temperatureThreshold/2)) { //Closed Intake Cooling Case 1 : abs(Target Temperature - Outside Temperature) <= (temperatureThreshold/2)
        Serial.println("Closed Intake Cooling Case 1 : abs(Target Temperature - Outside Temperature) <= (temperatureThreshold/2)");
        fanValue = 255; // Outside Temp is ideal temperature (Fan at max output)
      }
      else if(targetTemp < outsideTemp) {  //Closed Intake Cooling Case 2 : (Target Temperature < Outside Temperature)
        Serial.println("Closed Intake Cooling Case 2 : (Target Temperature < Outside Temperature)");
        fanValue =  mapFloat(abs(targetTemp - outsideTemp),0,10,1,255); // Higher outside temperature means higher fan output
      }
      else { //Closed Intake Cooling Case 3 : (Target Temperature > Outside Temperature)
        Serial.println("Closed Intake Cooling Case 3 : (Target Temperature > Outside Temperature)");
        fanValue = 255 - mapFloat(abs(targetTemp - outsideTemp),0,10,1,255); // Lower outside temperature means lower fan output 
      } 
    }
    Serial.println("Cooling Fan Value: ");
    Serial.println(fanValue);
    //analogWrite(heatingControlPin, 0); // Turn OFF resistive heating
    //analogWrite(fanControlPin, fanValue);
    Serial.println("The standard heating process will now be underway");
  }
}

//void UpdateSystem() {
//  outsideTemp = mapFloat(((analogRead(A0) - 20) * 3.04), 0, 1023, -40, 125)+32;
//  caseTemp = mapFloat(((analogRead(A1) - 20) * 3.04), 0, 1023, -40, 125)+32;
//  insideTemp = mapFloat(((analogRead(A2) - 20) * 3.04), 0, 1023, -40, 125)+32;
//
//  Serial.print("The outside temparature is now: ");
//  Serial.print(outsideTemp-32);
//  Serial.println(" ºC\n");
//
//  Serial.print("The case temparature is now: ");
//  Serial.print(caseTemp-32);
//  Serial.println(" ºC\n");
//
//  Serial.print("The inside temparature is now: ");
//  Serial.print(insideTemp-32);
//  Serial.println(" ºC\n");
//
//   /*Serial.println("oT" + String(outsideTemp-32));
//  Serial.println("cT" + String(caseTemp-32));
//  Serial.println("rT" + String(insideTemp-32));
//  Serial.println("tT" + String(targetTemp));*/
//}
float mapFloat(float value, float minValue, float maxValue, float toMinValue, float toMaxValue) {
  float mMinValue = (minValue > maxValue) ? maxValue : minValue;
  float mMaxValue = (minValue < maxValue) ? maxValue : minValue;
  float mToMinValue = (minValue > maxValue) ? toMaxValue : toMinValue;
  float mToMaxValue = (minValue < maxValue) ? toMaxValue : toMinValue;
  if(value <= mMinValue) return mToMinValue;
  else if(value >= mMaxValue) return mToMaxValue;
  return mToMinValue + ((mToMaxValue - mToMinValue) / (mMaxValue - mMinValue)) * (value - mMinValue);  
}

float getVal(float startingTemp, float targetTemp) {
  float energy = dryAirSpecificHeat * dryAirDensity * roomHeight * roomArea * (targetTemp - startingTemp);
  float power = energy/(heatingTime*60);
  Serial.println("POWER: " + String(power));
  if( (power/(float)numberResistance) < resistorRating*0.75) {
    float voltage = sqrt((resistanceValue * power)/numberResistance);
    if(voltage > 0 && voltage <= 5.0) return voltage;
    Serial.println("The heating process is not possible as the voltage required is too high.");
    Serial.println("The max voltage is 5V but the voltage required is " + String(voltage) + "V.");
    return -1;
  }
  Serial.println("The heating process is not possible with the wattage rating.");
  Serial.print("The power required per resistor is ");
  Serial.print((power/(float)numberResistance));
  Serial.print(" watt and the rating allowed is ");
  Serial.print((resistorRating*0.75));
  Serial.println(" watt.\n");
  return -1;
}

float getTargetTemp(float startingTemp, float voltage) {
  float power = (voltage*voltage*numberResistance)/(float)resistanceValue;
  Serial.println("POWER: " + String(power));
  float energy = power * heatingTime * 60;
  return (energy / (dryAirSpecificHeat * dryAirDensity * roomHeight * roomArea)) + startingTemp;
}

void buttonToggle() {
  if (!directionState){
    Serial.println("Closing Intake...\n");
    for(pos = 0; pos < 180; pos += 1) { // Close Intake
       servo.write(pos);
       delay(15);  // waits 15ms to reach the position
    }
  }
  else {
   Serial.println("Opening Intake...\n");
   for (pos = 180; pos>=1; pos -=1){ // Open Intake
     servo.write(pos);
     delay(15); // waits 15ms to reach the position
   }
  }
  directionState = !directionState; // Update Direction
  EEPROM.put(0, directionState); // Save Updated Direction to EEPROM of arduino
  buttonPressed = false;
}
